/*******************************************************************************
Copyright © 2023 Sebastian Roith

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*******************************************************************************/

/**
* Funktion, die die Datenbanktabellen und die Level erstellt
* @return { tables: "Array mit Datenbanktabellen", levels: "Array mit Leveldaten" }
*/
function getHumanDbmsLevels(
    fktRandomInt    // Funktion mit Paramtern min und max, die pseudozufällige aber wiedererzeugbare Folge von Ganzzahlen aus [min; max] liefert
) {
    /** Klasse für einfache Tabelle */
    function SimpleTable() {
        this.atts = [];
        this.data = [];
    }

    SimpleTable.prototype.addCol = function(name, values) {
        if (this.atts.indexOf(name) != -1)
            return;
        if (this.atts.length)
            this.data.forEach((value, index, table) => {
                if (value instanceof Array)
                    value.push(values[index]);
                else
                    table[index] = [value, values[index]];
            });
        else
            this.data = values.slice();
        this.atts.push(name);
    }

    SimpleTable.prototype.toString = function() {
        return [this.atts.join(';'), ...this.data.map(arr => arr.join(';'))].join('\n');
    }

    SimpleTable.prototype.clone = function() {
        const clone = new SimpleTable();
        clone.atts = this.atts.slice();
        clone.data = this.data.map(row => row.slice());
        return clone;
    }

    SimpleTable.prototype.select = function(atts) {
        const result = new SimpleTable();
        result.atts = atts;
        const indizes = atts.map(name => this.atts.indexOf(name));
        result.data = this.data.map(row => indizes.map(index => row[index]));
        return result;
    }

    SimpleTable.prototype.where = function(condition) {
        const result = new SimpleTable();
        result.atts = this.atts;
        result.data = this.data.filter(row => condition.exec(this.atts, row));
        return result;
    }

    SimpleTable.prototype.distinct = function() {
        const keep = this.data.map((row, index) => this.data.slice(0, index).reduce((prev, curr) => (prev && curr.reduce((p, c, i) => (p || (c != row[i])), false)), true));
        this.data = this.data.filter((row, index) => keep[index]);
        return keep;
    }

    SimpleTable.prototype.createComparer = function(name, descending) {
        const index = this.atts.indexOf(name);
        const sign = (descending ? -1 : 1);
        return function(a, b) {
            if (a[index] == b[index])
                return 0;
            if (typeof a[index] == "string" || typeof b[index] == "string")
                return sign * a[index].toString().localeCompare(b[index].toString());
            return sign * (a[index] - b[index]);
        }
    }

    SimpleTable.prototype.sort = function(rules) {
        const sortingFkts = rules.map(({ name, descending }) => this.createComparer(name, descending));
        let absoluteOrder = true;
        sortingFkts.push((a, b) => { absoluteOrder = false; return 0; });
        this.data.sort((a, b) => sortingFkts.reduce((prev, curr) => (prev || curr(a, b)), 0));
        if (absoluteOrder)
            return true;
        sortingFkts.pop();
        const groups = [];
        let lastGroup = [];
        this.data.forEach((row, index, arr) => {
            if (!index || sortingFkts.reduce((prev, curr) => (prev || curr(arr[index - 1], row)), 0)) {
                lastGroup = [ index ];
                groups.push(lastGroup);
            } else
                lastGroup.push(index);
        });
        return groups;
    }

    /** Eine ausführbare und in ein Level umwandelbare Abfrage */
    function SimpleQuery(
        tablename,          // Tabellenname
        table,              // SimpleTable-Instanz
        select,             // Optional: Array aus auszuwählenden Attributnamen (oder kein Array für *)
        distinct,           // Optional: Erneutes Auftreten gleicher Datensätze aussortieren?
        orderby,            // Optional: Sortiervorgaben der Form [ { name: "Attributname", descending: true }, { name: "Attributname 2" }, … ]
        where               // Optional: SimpleComparison-Instanz, die WHERE-Clause bildet
    ) {
        this.tablename = tablename;
        this.table = table;
        this.select = select;
        this.distinct = distinct;
        this.orderby = orderby;
        this.where = (where instanceof Array) ? new SimpleComparison(...where) : where;
    }

    SimpleQuery.prototype.toString = function() {
        let sql = '<strong>SELECT</strong> ';
        if (this.distinct)
            sql += '<strong>DISTINCT</strong> ';
        sql += (this.select instanceof Array) ? this.select.join(', ') : '*';
        sql += ' <strong>FROM</strong> ' + this.tablename;
        if (this.where)
            sql += ' <strong>WHERE</strong> ' + this.where.toString();
        if (this.orderby instanceof Array)
            sql += ' <strong>ORDER BY</strong> ' + this.orderby.map(({ name, descending}) => name + (descending ? ' <strong>DESC</strong>' : '')).join(', ');
        return sql;
    }

    SimpleQuery.prototype._exec = function() {
        let table = this.table.clone();
        const groups = (this.orderby instanceof Array) ? table.sort(this.orderby) : false;
        if (this.where)
            table = table.where(this.where);
        if (this.select instanceof Array)
            table = table.select(this.select);
        if (this.distinct) {
            const keepIndizes = [];
            table.distinct().forEach((keep, index) => { if (keep) keepIndizes.push(index); });
            if (groups instanceof Array)
                groups.forEach((group, index, arr) => (arr[index] = group.filter(ref => (keepIndizes.indexOf(ref) > -1)).map(oldIndex => keepIndizes.filter(i => (i < oldIndex)).length)));
        }
        return { table, groups: (groups instanceof Array) ? groups.filter(group => group.length) : groups };
    }

    SimpleQuery.prototype.toLevel = function() {
        const { table, groups } = this._exec();
        const level = {
            task: '<code>' + this.toString() + '</code>',
            locked: true,
            result: table.toString(),
            fine: 1
        };
        if (groups === true)
            level.orderRelevant = true;
        if (groups instanceof Array) {
            groups.forEach((group, index) => (groups[index] = group.map(rowIndex => table.data[rowIndex])));
            level.check = function({ data }) {
                const expected = groups.map(group => group.slice());
//console.log(expected);
                for (let i = 0; i < data.length; ++i) {
                    while (expected.length && !expected[0].length)
                        expected.shift();
                    if (!expected.length)
                        break;
                    const dataRow = data[i];
//console.log(dataRow, 'in', expected[0], '?');
                    const index = expected[0].reduce((prev, expectedRow, index) => (((prev == -1) && expectedRow.reduce((p, v, n) => (p && (v == dataRow[n])), true)) ? index : prev), -1);
                    if (index == -1)
                        return { msg: 'Sortierfehler: Datensatz ' + (i + 1) + ' darf erst später kommen', fine: 1 };
                    expected[0].splice(index, 1);
                }
                return { unlock: true };
            }
        }
//console.log(level.result);
        return level;
    }

    /** Ein Attributwert */
    function SimpleAttribute(
        name
    ) {
        this.name = name;
    }

    SimpleAttribute.prototype.toString = function() {
        if (typeof this.name == 'number')
            return '<em>' + this.name + '</em>';
        if (this.name.charAt(0) == "'")
            return '<em>' + this.name + '</em>';
        return this.name;
    }

    SimpleAttribute.prototype.exec = function(
        atts,
        row
    ) {
        if (typeof this.name == 'number')
            return this.name;
        if (this.name.charAt(0) == "'")
            return this.name.substring(1, this.name.length - 1);
        const index = atts.indexOf(this.name);
        if (index < 0)
            return undefined;
        return (row.length > index) ? row[index] : undefined;
    }

    /** Ein einfacher Vergleich oder eine algebraische Verknüpfung */
    function SimpleComparison(
        left,
        operator,
        right
    ) {
        if (this.operator != 'NOT')
            this.left = (typeof left == 'string') ? new SimpleAttribute(left) : ((left instanceof Array) ? new SimpleComparison(...left) : left);
        this.right = (typeof right == 'string') ? new SimpleAttribute(right) : ((right instanceof Array) ? new SimpleComparison(...right) : right);
        this.operator = operator;
    }
    SimpleComparison.prototype.left = undefined;
    SimpleComparison.prototype.right = undefined;

    SimpleComparison.specificy = [['*', '/'], ['+', '-'], ['<', '<=', '>', '>=', '<>', '='], ['NOT'], ['AND', 'OR']];
    SimpleComparison.getSpecificy = function(
        operator
    ) {
        const length = SimpleComparison.specificy.length;
        return length - SimpleComparison.specificy.reduce((prev, curr, index) => ((curr.indexOf(operator) > -1) ? index : prev), length + 1);
    }

    SimpleComparison.prototype.toString = function(
        parentOperator
    ) {
        const specificyThis = SimpleComparison.getSpecificy(this.operator);
        const specificyParent = SimpleComparison.getSpecificy(parentOperator);
        const needBrackets = (specificyThis < specificyParent) || ((specificyThis == specificyParent) && (parentOperator != this.operator));
        let s = '';
        if (this.left !== undefined)
            s = ((this.left instanceof SimpleComparison) ? this.left.toString(this.operator) : this.left) + ' ';
        s += '<strong>' + this.operator + '</strong> ';
        s += ((this.right instanceof SimpleComparison) ? this.right.toString(this.operator) : this.right);
        return needBrackets ? ('(' + s + ')') : s;
    }

    SimpleComparison.prototype.exec = function(
        atts,
        row
    ) {
        const left = this.left ? ((typeof this.left == 'number') ? this.left : this.left.exec(atts, row)) : false;
        const right = (typeof this.right == 'number') ? this.right : this.right.exec(atts, row);
        switch (this.operator) {
            case '=':
                return (left == right);
            case '<':
                return (left < right);
            case '<=':
                return (left <= right);
            case '>=':
                return (left >= right);
            case '>':
                return (left > right);
            case '<>':
                return (left != right);
            case '+':
                return (left + right);
            case '-':
                return (left - right);
            case '*':
                return (left * right);
            case '/':
                return (left / right);
            case 'NOT':
                return !right;
            case 'OR':
                return (left || right);
            case 'AND':
                return (left && right);
            default:
                return false;
        }
    }

    // Konkrete Tabellen zufällig erzeugen
    const availableChars = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ';

    function getUniqueValues(count, min, max) {
        const values = [];
        while (values.length < count) {
            const randomValue = fktRandomInt(min, max - values.length);
            values.push(values.slice().sort((a, b) => a - b).reduce((newValue, oldValue) => ((newValue >= oldValue) ? (newValue + 1) : newValue), randomValue));
        }
        return values;
    }

    function getValues(count, min, max) {
        const values = [];
        while (values.length < count)
            values.push(fktRandomInt(min, max));
        return values;
    }

    function copyValueOfCol(col, index1, index2) {
        const min = Math.min(index1, index2);
        const max = Math.max(index1, index2);
        col.splice(max, 0, col[min]);
        return col;
    }

    const result = {
        tabletype: 'Ergebnistabelle',
        tables: [
            {
                name: 'PUNKT',
                csv: 'name;x;y\n'
            },
            {
                name: 'STRECKE',
                csv: 'name;x1;y1;x2;y2\n'
            },
            {
                name: 'GERADE',
                csv: 'name;x1;y1;x2;y2\n'
            }
        ],
        levels: []
    };

    // Tabelle PUNKT
    const tablePunkt = new SimpleTable();
    const tablelengthPunkt = fktRandomInt(4, 5) + 2;
    tablePunkt.addCol('name', getUniqueValues(tablelengthPunkt, 0, availableChars.length - 1).map(index => availableChars.charAt(index)));
    const double = getUniqueValues(4, 0, tablelengthPunkt - 1);
    const indexXisY = double.pop();
    tablePunkt.addCol('x', copyValueOfCol(getUniqueValues(tablelengthPunkt - 1, 0, 2 * tablelengthPunkt), double[0], double[1]));
    tablePunkt.addCol('y', copyValueOfCol(getUniqueValues(tablelengthPunkt - 1, 0, 2 * tablelengthPunkt), double[2], double[1]));
    const valueXisY = tablePunkt.data[indexXisY][2];
    tablePunkt.data[indexXisY][1] = valueXisY;
    result.tables[0].csv = tablePunkt.toString();

    // Tabelle STRECKE
    const tableStrecke = new SimpleTable();
    const tablelengthStrecke = fktRandomInt(4, 5) + 2;
    tableStrecke.addCol('name', getUniqueValues(tablelengthStrecke, 0, availableChars.length - 1).map(index => availableChars.toLowerCase().charAt(index)));
    tableStrecke.addCol('x1', getUniqueValues(tablelengthStrecke, 0, 2 * tablelengthStrecke));
    tableStrecke.addCol('y1', getUniqueValues(tablelengthStrecke, 0, 2 * tablelengthStrecke));
    tableStrecke.addCol('x2', getUniqueValues(tablelengthStrecke, 0, 2 * tablelengthStrecke));
    tableStrecke.addCol('y2', getUniqueValues(tablelengthStrecke, 0, 2 * tablelengthStrecke));
    result.tables[1].csv = tableStrecke.toString();

    // Tabelle GERADE
    const tableGerade = new SimpleTable();
    const tablelengthGerade = 6 + 10 - tablelengthPunkt - tablelengthStrecke + 3;
    tableGerade.addCol('name', getUniqueValues(tablelengthGerade, 0, availableChars.length - 1).map(index => availableChars.toLowerCase().charAt(index)));
    tableGerade.addCol('x1', getValues(tablelengthGerade, 0, tablelengthGerade - 4));
    tableGerade.addCol('y1', getValues(tablelengthGerade, 0, tablelengthGerade - 4));
    tableGerade.addCol('x2', getValues(tablelengthGerade, 0, tablelengthGerade - 4));
    tableGerade.addCol('y2', getValues(tablelengthGerade, 0, tablelengthGerade - 4));
    const xIsY = getValues(2, 0, tablelengthGerade - 4);
    const indexGeradeXisY = fktRandomInt(0, tablelengthGerade - 1);
    tableGerade.data[indexGeradeXisY][1] = xIsY[0];
    tableGerade.data[indexGeradeXisY][2] = xIsY[0];
    tableGerade.data[indexGeradeXisY][3] = xIsY[1];
    tableGerade.data[indexGeradeXisY][4] = xIsY[1];
    result.tables[2].csv = tableGerade.toString();

    // Level erstellen
    result.levels.push((new SimpleQuery('PUNKT', tablePunkt, '*', false, undefined, new SimpleComparison(undefined, 'NOT', new SimpleComparison('name', '=', "'" + tablePunkt.data[fktRandomInt(1, tablelengthPunkt - 2)][0] + "'")))).toLevel());
    result.levels.push((new SimpleQuery('STRECKE', tableStrecke, ['y1'], false, undefined, [
        ['y1', '>', tablelengthStrecke + Math.floor(tablelengthStrecke / 3)],
        'OR',
        ['y1', '<', tablelengthStrecke - Math.floor(tablelengthStrecke / 3)]
    ])).toLevel());
    result.levels.push((new SimpleQuery('GERADE', tableGerade, '*', false, undefined, [
        undefined,
        'NOT',
        [
            ['x1', '=', 'y1'],
            'AND',
            ['x2', '=', 'y2']
        ]
    ])).toLevel());
    result.levels.push((new SimpleQuery('PUNKT', tablePunkt, ['x', 'y'], false, undefined, [['x', '<>', 'y'], 'AND', (valueXisY * 4 > tablelengthPunkt) ? ['x', '<', tablelengthPunkt + 1] : ['x', '>', tablelengthPunkt - 1]])).toLevel());
    result.levels.push((new SimpleQuery('GERADE', tableGerade, ['name'], false, undefined, [
        ['x1', '=', xIsY[0]],
        'OR',
        [
            undefined,
            'NOT',
            ['x2', '=', xIsY[1]]
        ]
    ])).toLevel());
    result.levels.push((new SimpleQuery('STRECKE', tableStrecke, '*', false, undefined, [
        ['name', '>', "'k'"],
        'AND',
        [
            ['x1', '>=', Math.floor(2 * tablelengthStrecke / 3)],
            'AND',
            ['x2', '<', 2 * tablelengthPunkt - Math.floor(3 * tablelengthStrecke / 4)]
        ]
    ])).toLevel());
    result.levels.push((new SimpleQuery('STRECKE', tableStrecke, '*', false, undefined, [
        ['y2', '>', 2 * tablelengthPunkt - Math.floor(tablelengthStrecke / 3)],
        'OR',
        [
            ['x2', '<=', Math.floor(tablelengthStrecke / 3)],
            'OR',
            ['x1', '>', 2 * tablelengthPunkt - Math.floor(tablelengthStrecke / 4)]
        ]
    ])).toLevel());
    result.levels.push((new SimpleQuery('PUNKT', tablePunkt, ['name'], false, undefined, [
        ['x', '<', tablePunkt.data.reduce((prev, row) => Math.max(prev, row[1]), 0) - 1],
        'OR',
        ['y', '>=', tablePunkt.data.reduce((prev, row) => Math.min(prev, row[2]), Infinity) + 2],
    ])).toLevel());
    result.levels.push((new SimpleQuery('GERADE', tableGerade, '*', false, undefined, [
        [undefined, 'NOT', ['x1', '=', 'y1']],
        'OR',
        ['x2', '=', 0]
    ])).toLevel());
    result.levels.push((new SimpleQuery('STRECKE', tableStrecke, '*', false, undefined, [
        ['name', '<', "'m'"],
        'OR',
        [
            ['x1', '>=', Math.floor(tablelengthStrecke / 3)],
            'AND',
            ['x2', '<', 2 * tablelengthPunkt - Math.floor(tablelengthStrecke / 4)]
        ]
    ])).toLevel());

    return result;
}
